<?php
namespace Zeedhi\Tools\DatabaseReader\Adapters;

use Doctrine\DBAL\Connection;
use Zeedhi\Tools\DatabaseReader\AdapterReader;
use Zeedhi\Tools\DatabaseReader\DTO;
use Zeedhi\Tools\DatabaseReader\Exception;

class MySQL extends AdapterReader{

    protected static $deleteRuleConversion = array(
        'CASCADE'   => DTO\Constraints\ForeignKey::DELETE_RULE_CASCADE,
        'NO ACTION' => DTO\Constraints\ForeignKey::DELETE_RULE_NO_ACTION,
        'RESTRICT'  => DTO\Constraints\ForeignKey::DELETE_RULE_RESTRICT,
        'SET NULL'  => DTO\Constraints\ForeignKey::DELETE_RULE_SET_NULL
    );

    const CONSTRAINT_TYPE_PK = "PRIMARY";
    const CONSTRAINT_TYPE_FK = "FOREIGN";
    const CONSTRAINT_TYPE_UNIQUE = "UNIQUE";

    protected static $columnDataTypeConversion = array(
        'binary'            => DTO\Column::DATA_TYPE_BINARY,
        'tinyblob'          => DTO\Column::DATA_TYPE_BINARY,
        'blob'              => DTO\Column::DATA_TYPE_BINARY,
        'mediumblob'        => DTO\Column::DATA_TYPE_BINARY,
        'longblob'          => DTO\Column::DATA_TYPE_BINARY,

        'bool'              => DTO\Column::DATA_TYPE_BOOLEAN,
        'boolean'           => DTO\Column::DATA_TYPE_BOOLEAN,

        'bit'               => DTO\Column::DATA_TYPE_INTEGER,
        'tinyint'           => DTO\Column::DATA_TYPE_INTEGER,
        'smallint'          => DTO\Column::DATA_TYPE_INTEGER,
        'mediumint'         => DTO\Column::DATA_TYPE_INTEGER,
        'int'               => DTO\Column::DATA_TYPE_INTEGER,
        'integer'           => DTO\Column::DATA_TYPE_INTEGER,
        'bigint'            => DTO\Column::DATA_TYPE_INTEGER,
        'serial'            => DTO\Column::DATA_TYPE_INTEGER,

        'number'            => DTO\Column::DATA_TYPE_FLOAT,
        'decimal'           => DTO\Column::DATA_TYPE_FLOAT,
        'dec'               => DTO\Column::DATA_TYPE_FLOAT,
        'float'             => DTO\Column::DATA_TYPE_FLOAT,
        'double'            => DTO\Column::DATA_TYPE_FLOAT,
        'double precision'  => DTO\Column::DATA_TYPE_FLOAT,

        'varchar'           => DTO\Column::DATA_TYPE_STRING,
        'varbinary'         => DTO\Column::DATA_TYPE_STRING,
        'enum'              => DTO\Column::DATA_TYPE_STRING,
        'set'               => DTO\Column::DATA_TYPE_STRING,

        'text'              => DTO\Column::DATA_TYPE_STRING,
        'mediumtext'        => DTO\Column::DATA_TYPE_STRING,
        'longtext'          => DTO\Column::DATA_TYPE_STRING,

        'char'              => DTO\Column::DATA_TYPE_STRING,
        'char byte'         => DTO\Column::DATA_TYPE_STRING,
        'character'         => DTO\Column::DATA_TYPE_STRING,
        'ascii'             => DTO\Column::DATA_TYPE_STRING,
        'unicode'           => DTO\Column::DATA_TYPE_STRING,

        'date'              => DTO\Column::DATA_TYPE_DATE,
        'datetime'          => DTO\Column::DATA_TYPE_DATETIME,
        'timestamp'         => DTO\Column::DATA_TYPE_DATETIME,
        'time'              => DTO\Column::DATA_TYPE_DATETIME,
        'year'              => DTO\Column::DATA_TYPE_DATE
    );

    /** @var \Doctrine\DBAL\Connection */
    protected $connection;

    public function __construct(Connection $connection) {
        $this->connection = $connection;
    }

    /**
     * Get name from all tables in database.
     *
     * @return string[] With name of all tables.
     */
    public function getAllTableNames() {
        $tableRows = $this->connection->fetchAll("SHOW TABLES");
        $tableNames = array();
        foreach($tableRows as $key => $dbTableArray){
            foreach($dbTableArray as $dbTableName){
                $tableNames[] = $dbTableName;
            }
        }

        return $tableNames;
    }

    /**
     * Get a table DTObject with all data about chosen table.
     *
     * @param string $tableName Name from chosen table.
     *
     * @return DTO\Table
     *
     * @throws \Zeedhi\Tools\DatabaseReader\Exception When table was not found in database.
     */
    public function getTable($tableName) {
        $tableRow = $this->getTableRow($tableName);
        $DTOTable = new DTO\Table();
        $DTOTable->setName($tableRow['TABLE_NAME']);
        return $DTOTable;
    }

    /**
     * Return constraints rows of given table that belongs to chosen table.
     *
     * @param string $tableName      Name of table that belongs constraints.
     * @param string $constraintType Type of constraint, must be on of self::CONSTRAINT_TYPE_*.
     *
     * @return array of Rows that are array too.
     */
    protected function getConstraintRow($tableName, $constraintType) {
        $statement = $this->connection->prepare("SHOW KEYS FROM $tableName WHERE KEY_NAME='$constraintType'");
        $statement->execute();
        return $statement->fetchAll();
    }

    protected function getPrimaryKeyRow($tableName) {
        $primaryKeyRow = $this->getConstraintRow($tableName, self::CONSTRAINT_TYPE_PK);
        if (!$primaryKeyRow) throw Exception::tableWithoutPrimaryKey($tableName);
        return $primaryKeyRow;
    }

    protected function getForeignKeyRows($tableName) {
        $statement = $this->connection->prepare("SELECT DISTINCT k.CONSTRAINT_NAME, k.REFERENCED_COLUMN_NAME, c.DELETE_RULE
                                                                FROM information_schema.key_column_usage k
                                                                INNER JOIN information_schema.referential_constraints c ON
                                                                c.constraint_name = k.constraint_name AND
                                                                c.table_name = '$tableName' WHERE k.table_name = '$tableName'");
        $statement->execute();
        return $statement->fetchAll();
    }

    protected function getUniqueConstraintRows($tableName) {
        return $this->getConstraintRow($tableName, self::CONSTRAINT_TYPE_UNIQUE);
    }

    protected function getDTOForeignKeyFromRow($foreignKeyRow){
        if(!isset(self::$deleteRuleConversion[$foreignKeyRow['DELETE_RULE']])) {
            throw Exception::unsupportedDeleteRule($foreignKeyRow['DELETE_RULE']);
        }
        $deleteRule = self::$deleteRuleConversion[$foreignKeyRow['DELETE_RULE']];
        $DTOForeignKey = new DTO\Constraints\ForeignKey(
            $foreignKeyRow['CONSTRAINT_NAME'],
            $foreignKeyRow['REFERENCED_COLUMN_NAME'],
            $deleteRule
        );
        return $DTOForeignKey;
    }

    protected function getDTOUniqueConstraintsFromRows($uniqueConstraintRows) {
        $DTOUniqueConstraint = array();
        foreach($uniqueConstraintRows as $uniqueConstraintRow) {
            $DTOUniqueConstraint[] = new DTO\Constraints\Unique($uniqueConstraintRow['Column_name']);
        }

        return $DTOUniqueConstraint;
    }

    /**
     * @param $tableName
     *
     * @return mixed
     *
     * @throws \Zeedhi\Tools\DatabaseReader\Exception
     */
    protected function getTableRow($tableName) {
        static $statement = null;
        if ($statement === null) {
            $statement = $this->connection->prepare("SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME=:TABLE_NAME");
        }

        $statement->execute(array('TABLE_NAME' => $tableName));
        $tableRow = $statement->fetch(\PDO::FETCH_ASSOC);
        if (!$tableRow) throw Exception::tableNotFound($tableName);
        return $tableRow;
    }

    /**
     * @param $tableName
     * @return array
     */
    protected function getColumnRows($tableName) {
        static $statement = null;
        if ($statement === null) {
            $statement = $this->connection->prepare("SELECT COLUMN_NAME, DATA_TYPE, CHARACTER_MAXIMUM_LENGTH, NUMERIC_PRECISION, NUMERIC_SCALE, IS_NULLABLE, COLUMN_DEFAULT FROM
                                                     INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME = :TABLE_NAME");
        }

        $statement->execute(array('TABLE_NAME' => $tableName));
        return $statement->fetchAll();
    }

    /**
     * Factory a DTO\Column from a row of MySQL Dictionary.
     *
     * @param array $columnRow With positions COLUMN_NAME, DATA_DEFAULT, DATA_LENGTH, DATA_PRECISION, DATA_SCALE, DATA_TYPE, NULLABLE
     *
     * @return DTO\Column
     */
    protected function getDTOColumnFromRow($columnRow) {
        $DTOColumn = new DTO\Column();
        $DTOColumn->setName($columnRow['COLUMN_NAME']);
        $DTOColumn->setDataType(self::$columnDataTypeConversion[$columnRow['DATA_TYPE']]);
        $DTOColumn->setDataDefault($columnRow['COLUMN_DEFAULT']);
        $DTOColumn->setDataLength($columnRow['CHARACTER_MAXIMUM_LENGTH']);
        $DTOColumn->setDataPrecision($columnRow['NUMERIC_PRECISION']);
        $DTOColumn->setDataScale($columnRow['NUMERIC_SCALE']);
        $DTOColumn->setNullAble($columnRow['IS_NULLABLE'][0]);
        return $DTOColumn;
    }

    /**
     * @param $primaryKeyRow
     * @return DTO\Constraints\PrimaryKey
     */
    protected function getDTOPrimaryKeyFromRow($primaryKeyRow) {
        $DTOPrimaryKey = new DTO\Constraints\PrimaryKey($primaryKeyRow['Column_name']);
        return $DTOPrimaryKey;
    }

    protected function getDTOConsColumnFromRow($consColumnRow) {
        $DTOConsColumn = new DTO\Constraints\ConsColumn($consColumnRow['COLUMN_NAME'], $consColumnRow['ORDINAL_POSITION']);
        return $DTOConsColumn;
    }

    protected function getConsColumnsRow($constraintName) {
        static $statement = null;
        if ($statement === null) {
            $statement = $this->connection->prepare("SELECT COLUMN_NAME, ORDINAL_POSITION FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE WHERE CONSTRAINT_NAME=:CONSTRAINT_NAME");
        }

        $statement->execute(array(":CONSTRAINT_NAME" => $constraintName));
        return $statement->fetchAll();
    }

    /**
     * Get foreign key from chosen table with given name.
     *
     * @param string $tableName Name from chosen table.
     * @param string $constraintName Name of constraint.
     *
     * @throws \Zeedhi\Tools\DatabaseReader\Exception When table was not found in database.
     *
     * @return DTO\Constraints\ForeignKey
     */
    public function getForeignKeyByNameAndTableName($constraintName, $tableName){
        $statement = $this->connection->prepare("SELECT DISTINCT k.CONSTRAINT_NAME, k.REFERENCED_COLUMN_NAME, c.DELETE_RULE
                                                                FROM information_schema.key_column_usage k
                                                                INNER JOIN information_schema.referential_constraints c ON
                                                                c.constraint_name = k.constraint_name AND
                                                                c.table_name = :TABLE_NAME WHERE k.table_name = :TABLE_NAME AND k.CONSTRAINT_NAME = :CONSTRAINT_NAME");
        $statement->execute(array(':TABLE_NAME' => $tableName, ':CONSTRAINT_NAME' => $constraintName));
        $constraintRow = $statement->fetch(\PDO::FETCH_ASSOC);
        if (!$constraintRow) throw Exception::constraintNotFoundInDatabase($tableName, $constraintName);
        $DTOForeignKey = $this->getDTOForeignKeyFromRow($constraintRow);
        $this->completeConsColumnForConstraint($DTOForeignKey);
        return $DTOForeignKey;
    }

    /**
     * Get the sequence of that table.
     *
     * @param string $tableName Name of table.
     * @param string $columnName Name of column that uses that sequence.
     *
     * @return DTO\Sequence
     */
    public function getSequenceFromTableAndColumn($tableName, $columnName){
        $statement = $this->connection->prepare("SELECT * FROM INFORMATION_SCHEMA.COLUMNS WHERE EXTRA like '%auto_increment%'
                                                AND TABLE_NAME = :TABLE_NAME AND COLUMN_NAME = :COLUMN_NAME");
        $statement->execute(array(':TABLE_NAME' => $tableName, ':COLUMN_NAME' => $columnName));
        if(!$statement->fetch(\PDO::FETCH_ASSOC)){
            throw Exception::sequenceDoesNotExist($tableName, $columnName);
        }
        return new DTO\Sequence("seq_".$tableName."_".$columnName);
    }


}